# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import pytest
import sqlalchemy

from gaarf import sql_executor


class TestSqlAlchemyQueryExecutor:
  @pytest.fixture
  def engine(self):
    return sqlalchemy.create_engine('sqlite:///:memory:')

  def test_instantiating_sql_executor_is_deprecated(self, engine):
    with pytest.warns(DeprecationWarning) as w:
      sql_executor.SqlAlchemyQueryExecutor(engine)
      assert len(w) == 1
      assert issubclass(w[0].category, DeprecationWarning)
